from . import log
logger = log.get(__name__)

import os
from .utils import delay
import random as rand
import time

class Executor:
    
    def __init__(self, peerAccessor, policy, zipkin, target, trackerMethod, binary_annotations, actionCb):
        self._peerAccessor = peerAccessor
        self._policy = policy
        self._target = target
        self._trackerMethod = trackerMethod
        self._binary_annotations = binary_annotations
        self._actionCb = actionCb
        self._zipkin = None
        
        if zipkin.isEnabled():
            self._zipkin = zipkin

        if target[0] == 'service':
            self._remoteServiceName = os.environ['BERLIOZ_CLUSTER'] + '-' + target[1]
        elif target[0] == 'cluster':
            self._remoteServiceName = target[1] + '-' + target[2]
        else:
            self._remoteServiceName = os.environ['BERLIOZ_CLUSTER'] + '-' + '-'.join(target)

        self._context = {
            "canRetry": True,
            "tryCount": 0
        }

    def perform(self):
        while self._canTry():
            self._retryWait()
            self._try()
            if self._checkCompleted():
                return self._context['result']
        exInfo = self._context['lastError']
        if exInfo:
            raise exInfo

    def _canTry(self):
        if self._context['tryCount'] == 0:
            return True
        if self._context['tryCount'] >= self._resolvePolicy('retry-count'):
            return False
        return self._context['canRetry']

    def _try(self):
        logger.info('_try begin')

        self._context['hasError'] = False
        self._context['lastError'] = None
        self._context['tryCount'] = self._context['tryCount'] + 1

        self._perform()
        logger.info('_try end')

    def _retryWait(self):
        if self._context['tryCount'] == 0:
            return
        logger.info('_retryWait')

        timeout = self._resolvePolicy('retry-initial-delay')
        timeout = timeout * pow(self._resolvePolicy('retry-delay-multiplier'), self._context['tryCount'] - 1)
        timeout = min(timeout, self._resolvePolicy('retry-max-delay'))
        logger.info('_retryWait timeout: %s', timeout)

        if timeout > 0:
            # tracer = this._instrument('sleep', 'GET', 'http://sleep')
            time.sleep(timeout / 1000)

    def _perform(self):
        try:
            peer = self._fetchPeer()
            if not peer:
                if not self._resolvePolicy('no-peer-retry'):
                    self._context['canRetry'] = False
                raise Exception('No peers found', self._target)

            zipki_span = None
            if self._zipkin:
                zipki_span = self._zipkin.instrumentRequest('-'.join(self._target),
                                                            self._trackerMethod,
                                                            self._binary_annotations)
            if zipki_span:
                with zipki_span:
                    result = self._actionCb(peer, zipki_span)
            else:
                result = self._actionCb(peer)
            logger.debug('Result: %s', result)
            self._context['result'] = result
        except Exception as ex:
            self._context['hasError'] = True
            self._context['lastError'] = ex
            logger.exception('operation failed')

        # var tracer = this._instrument(this._remoteServiceName, this._trackerMethod, this._trackerUrl);
        # return Promise.resolve()
        #     .then(result => {
        #         tracer.finish(200);
        #         this._context.result = result;
        #     })
        #     .catch(reason => {
        #         tracer.error(reason);
        #         this._context.hasError = true;
        #         this._context.lastError = reason;
        #         this._logger.error('Executor::_perform::catch: ', reason);
        #     });
    
    def _fetchPeer(self):
        peer = self._peerAccessor.random()
        return peer

    def _checkCompleted(self):
        if self._context['hasError']:
            return False
        return True

    def _resolvePolicy(self, name):
        return self._policy.resolve(name, self._target)


from . import log
logger = log.get(__name__)

from . import metadata
logger.info('BerliozPythonSDK v%s', metadata.VERSION)

import os
if not os.environ.get('BERLIOZ_CLUSTER'):
    logger.warning('Using berlioz sdk outside of managed environment.')
else:
    from . import starter as _starter
    from .secret import SecretClient
    from .cluster import Cluster
    from .sector import Sector

    from functools import wraps
    import inspect

    # PEERS
    def cluster(name, endpoint=None):
        return Cluster(_starter, name, endpoint)

    def sector(name):
        return Sector(_starter, name)

    def service(name, endpoint=None):
        return sector(_starter.berlioz_sector).service(name, endpoint)

    def database(name):
        return sector(_starter.berlioz_sector).database(name)

    def queue(name):
        return sector(_starter.berlioz_sector).queue(name)


    # # SECRET PUBLIC KEY
    # def monitorSecretPublicKey(name, cb):
    #     _starter.monitorNatives('secret_public_key', name, cb)

    # def getSecretPublicKeys(name):
    #     return _starter.getNatives('secret_public_key', name)

    # def getSecretPublicKey(name):
    #     return _starter.getNative('secret_public_key', name)

    # def getSecretPublicKeyClient(name):
    #     return _starter.getNativeClient('secret_public_key', name)

    # # SECRET PRIVATE KEY
    # def monitorSecretPrivateKey(name, cb):
    #     _starter.monitorNatives('secret_private_key', name, cb)

    # def getSecretPrivateKeys(name):
    #     return _starter.getNatives('secret_private_key', name)

    # def getSecretPrivateKey(name):
    #     return _starter.getNative('secret_private_key', name)

    # def getSecretPrivateKeyClient(name):
    #     return _starter.getNativeClient('secret_private_key', name)

    # # SECRET PUBLIC & PRIVATE KEY
    # def getSecret(name):
    #     return SecretClient(_starter, name)

    # TRACING
    def instrument(method, binary_annotations=None):
        return _starter.instrument(method, binary_annotations)

    def instrument_func(**instArgs):
        def decorator(f):
            @wraps(f)
            def wrapper(*args, **kwargs):
                mod = inspect.getmodule(f)
                binary_annotations = instArgs.get('binary_annotations', {})
                binary_annotations['source.path'] = mod.__file__
                binary_annotations['source.func'] = f.__name__
                binary_annotations['source.module'] = mod.__name__
                method = instArgs.get('name')
                if not method:
                    method = 'func-' + f.__name__

                with instrument(method, binary_annotations):
                    return f(*args, **kwargs)
            return wrapper
        return decorator

    # FRAMEWORKS
    def setupFlask(app):
        return _starter.setupFlask(app)


from . import log
logger = log.get(__name__)

import json

class Registry:
    
    def __init__(self):
        self._sections = {}
        self._subscribers = {}

    def extractRoot(self):
        return self._sections

    def set(self, sectionName, path, value):
        logger.debug('SET. Section: %s, Path: %s, Value: %s', sectionName, path, value)

        section = self._getSection(sectionName)
        pathStr = self._getKey(path)
        # currValue = section.get(pathStr)
        # if (_.isEqual(currValue, value)) {
        #     return;
        # }
        section[pathStr] = value
        self._notifyToSubscribers(sectionName, path)

    def get(self, sectionName, path):
        section = self._getSection(sectionName)
        pathStr = self._getKey(path)
        return section.get(pathStr)

    def reset(self, sectionName):
        logger.debug('RESET. Section: %s', sectionName)
        self._sections[sectionName] = {}

    def subscribe(self, sectionName, path, cb):
        subscriberId = self._getSubscriberId(sectionName, path)
        if not subscriberId in self._subscribers:
            self._subscribers[subscriberId] = []
        self._subscribers[subscriberId].append(cb)
        self._notifyToSubscriber(sectionName, path, cb)

    def _getKey(self, path):
        return json.dumps(path)

    def _getSection(self, name):
        if not name in self._sections:
            self._sections[name] = {}
        return self._sections[name]

    def _notifyToSubscribers(self, sectionName, path):
        subscriberId = self._getSubscriberId(sectionName, path)
        if not subscriberId in self._subscribers:
            return
        value = self.get(sectionName, path)
        if not value:
            return
        for cb in self._subscribers[subscriberId]:
            self._triggerToSubscriber(value, cb)

    def _notifyToSubscriber(self, sectionName, path, cb):
        value = self.get(sectionName, path)
        if not value:
            return
        self._triggerToSubscriber(value, cb)

    def _triggerToSubscriber(self, value, cb):
        try:
            cb(value)
        except Exception:
            logger.exception('Caught from subscriber')

    def _getSubscriberId(self, sectionName, path):
        subscriber = {
            "section": sectionName,
            "path": path
        }
        return json.dumps(subscriber)



from . import log
logger = log.get(__name__)

from .utils import delay
import os
import websocket
import json
import threading

class Client(): 

    def __init__(self, handler): 
        logger.info('Constructing.')
        self._handler = handler
        self.ws = None
        self.isClosed = False
        websocket.enableTrace(False)
        self._connect()

    def close(self):
        self.isClosed = True
        if self.ws is not None:
            self.ws.close()

    def _connect(self):
        if self.ws is not None:
            return
        if self.isClosed:
            return

        logger.info('Connecting...')
        ws = websocket.WebSocketApp(os.environ['BERLIOZ_AGENT_PATH'],
                                    on_message = self._onMessage,
                                    on_error = self._onError,
                                    on_close = self._onClose)
        ws.on_open = self._onOpen
        ws_thread = threading.Thread(target=ws.run_forever)
        ws_thread.daemon = True
        self.ws = ws
        ws_thread.start()

    def _onMessage(self, ws, message):
        if not ws is self.ws:
            return
        logger.info('Message received.')
        logger.debug('Message RAW contents: %s', message)
        data = json.loads(message)
        logger.debug('Message JSON Contents: %s', data)
        self._handler(data)
        logger.info('Message processed.')

    def _onError(self, ws, error):
        logger.error('Error')
        logger.error(error)
        ws.close()

    def _onClose(self, ws):
        logger.info('Closed')
        self.ws = None
        self._reconnect()

    def _onOpen(self, ws):
        logger.info('Opened')

    def _reconnect(self):
        if self.isClosed:
            return
        delay(1000, self._connect)



from flask import g as flask_g
from flask import request as flask_request
from flask import Response
from flask import json
import pprint

class Flask:
    def __init__(self, app, zipkin, policy, registry):
        self._app = app
        self._zipkin = zipkin
        self._policy = policy
        self._registry = registry
        
        self._app.before_request(self._handleBeforeRequest)
        self._app.after_request(self._handleAfterRequest)
        
        self._app.add_url_rule('/berlioz', 'berlioz_debug', view_func=self._berlioz_debug)


    def _handleBeforeRequest(self):
        url = ''
        if flask_request.script_root is not None:
           url = flask_request.script_root
        if flask_request.path is not None:
           url = url + flask_request.path
        zipkin_span = self._zipkin.instrumentServer(flask_request.headers,
                                                    flask_request.method,
                                                    url)
        flask_g._berlioz_zipkin_span = zipkin_span
        zipkin_span.start()

    def _handleAfterRequest(self, response):
        zipkin_span = getattr(flask_g, '_berlioz_zipkin_span')
        if zipkin_span:
            self._zipkin.serverResponse(zipkin_span, response.status_code)
        return response

    def _berlioz_debug(self):
        data = pprint.pformat(self._registry.extractRoot())
        html = '<html><body><pre>' + data + '</pre></body></html>'
        return html


from . import log
logger = log.get(__name__)

import requests
import os
import platform

if platform.system() == 'Windows':
    import win_inet_pton

from py_zipkin.zipkin import zipkin_span
from py_zipkin.zipkin import ZipkinAttrs
from py_zipkin.util import generate_random_64bit_string as zipkin_generate_span
from py_zipkin.storage import ThreadLocalStack as ZipkinThreadLocalStack

class Zipkin:
    
    def __init__(self, peerHelper, policy):
        self._peerHelper = peerHelper
        self._policy = policy
        self._zipkin_context_stack = ZipkinThreadLocalStack()
        self._localName = 'service://' + os.environ['BERLIOZ_CLUSTER'] + '-' + os.environ['BERLIOZ_SECTOR'] + '-' + os.environ['BERLIOZ_SERVICE']
        self._sampleRate = 100
        self._endpoint = None
        self._policy.monitor('enable-zipkin', [], self._onZipkinEnabledChanged)
        self._policy.monitor('zipkin-service-id', [], self._onZipkinServiceIdChanged)

    def isEnabled(self):
        return self._isEnabled and self._endpoint

    def addZipkinHeaders(self, headers, zipkin_span):
        if not zipkin_span:
            return
        zipkin_attrs = zipkin_span.zipkin_attrs
        headers['X-B3-TraceId'] = zipkin_attrs.trace_id
        headers['X-B3-SpanId'] = zipkin_attrs.span_id
        headers['X-B3-ParentSpanId'] = zipkin_attrs.parent_span_id
        headers['X-B3-Flags'] = zipkin_attrs.flags
        headers['X-B3-Sampled'] = str(int(zipkin_attrs.is_sampled))

    def instrument(self, spanName, binary_annotations=None):
        zipkin_attrs = None
        existing_zipkin_attrs = self._zipkin_context_stack.get()
        if existing_zipkin_attrs:
            zipkin_attrs = ZipkinAttrs(
                span_id = zipkin_generate_span(),
                trace_id = existing_zipkin_attrs.trace_id,
                parent_span_id = existing_zipkin_attrs.span_id,
                flags = existing_zipkin_attrs.flags,
                is_sampled = existing_zipkin_attrs.is_sampled
            )
        return self._instrument(self._localName, spanName, zipkin_attrs, binary_annotations)

    def instrumentRequest(self, remoteServiceName, method, binary_annotations=None):
        zipkin_attrs = None
        existing_zipkin_attrs = self._zipkin_context_stack.get()
        if existing_zipkin_attrs:
            zipkin_attrs = ZipkinAttrs(
                span_id = zipkin_generate_span(),
                trace_id = existing_zipkin_attrs.trace_id,
                parent_span_id = existing_zipkin_attrs.span_id,
                flags = existing_zipkin_attrs.flags,
                is_sampled = existing_zipkin_attrs.is_sampled
            )
        return self._instrument(remoteServiceName, method, zipkin_attrs, binary_annotations)

    def instrumentServer(self, headers, method, url):
        # See https://github.com/openzipkin/zipkin-js/blob/7588d88b876ab5f5844f4bbd08f50c16c13306ab/packages/zipkin/src/instrumentation/httpServer.js
        zipkin_attrs = ZipkinAttrs(
            span_id = headers.get('X-B3-SpanId') or zipkin_generate_span(),
            trace_id = headers.get('X-B3-TraceId') or zipkin_generate_span(),
            parent_span_id = headers.get('X-B3-ParentSpanId'),
            flags = headers.get('X-B3-Flags'),
            is_sampled = str(headers.get('X-B3-Sampled') or '0') == '1'
        )
        binary_annotations = {
            'http.url': url
        }
        return self._instrument(self._localName, method, zipkin_attrs, binary_annotations)

    def serverResponse(self, zipkin_span, status_code):
        zipkin_span.update_binary_annotations({
            'http.status_code': status_code 
        }) 
        zipkin_span.stop()

    def _instrument(self, serviceName, spanName, zipkin_attrs, binary_annotations=None):
        zipArgs = {
            "service_name": serviceName,
            "span_name": spanName,
            "transport_handler": self._zipkin_http_transport,
            "port": 0,
            "sample_rate": self._sampleRate,
            "zipkin_attrs": zipkin_attrs,
            "binary_annotations": binary_annotations
        }
        return zipkin_span(**zipArgs)

    def _zipkin_http_transport(self, encoded_span):
        if not self._isEnabled:
            return
        zipkin_url = self._endpoint
        if not zipkin_url:
            return
        # TODO: check https://github.com/Yelp/py_zipkin#transport for
        # body = b"\x0c\x00\x00\x00\x01" + encoded_span
        headers = {"Content-Type": "application/x-thrift"}
        # TODO: Run in a thread
        try:
            requests.post(zipkin_url, data=encoded_span, headers=headers)
        except Exception as e:
            logger.error(e)

    def _onZipkinEnabledChanged(self, value):
        logger.info('ZipkinEnabled = %s', value)
        self._isEnabled = value

    def _onZipkinServiceIdChanged(self, value):
        logger.info('ZipkinServiceId = %s', value)
        peerPath = [value, 'client']
        def onPeerchanged(peer):
            self._endpoint = peer['protocol'] + '://' + peer['address'] + ':' + str(peer['port']) + '/api/v1/spans'
            logger.info('ZipkinServiceEndpoint = %s', self._endpoint)
        self._peerHelper.monitorPeer(peerPath, self._peerHelper.selectFirstPeer, onPeerchanged)

